#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / tasks / package
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'task_hub',
        ]

from ..helpers.hub import BaseHub
from ..helpers.modprober import ModProber
from ..registry.classes import UnicodeRegistry

TASK_DRIVERS_KEY = 'drivers'
APP_OBJECTS_KEY = 'app_objects'
APP_CONF_KEY = 'conf'

PROBER = ModProber('weiyu.tasks', '%sdriver')


class TaskHub(BaseHub):
    registry_name = 'weiyu.tasks'
    registry_class = UnicodeRegistry
    handlers_key = TASK_DRIVERS_KEY

    def __init__(self):
        super(TaskHub, self).__init__()
        self._apps = self._reg[APP_OBJECTS_KEY] = {}

    def ensure_app(self, name):
        if name in self._apps:
            return

        # Read app configuration.
        dct = self._reg[APP_CONF_KEY][name]
        drv, conf, extras = dct['driver'], dct['conf'], dct['extras']

        # Load driver.
        PROBER.modprobe(drv)

        # Build app object.
        app_obj = self.do_handling(drv).build_app(conf, extras)

        # Save a reference, along with the driver type
        self._apps[name] = {
                'drv': drv,
                'obj': app_obj,
                }

    def get_app(self, name):
        try:
            return self._apps[name]['obj']
        except KeyError:
            self.ensure_app(name)
            return self._apps[name]['obj']

    def task(self, app):
        # Callback into driver layer for high-level API consistence
        try:
            appreg = self._apps[app]
        except KeyError:
            self.ensure_app(app)
            appreg = self._apps[app]

        return self.do_handling(appreg['drv']).get_task(appreg['obj'])


task_hub = TaskHub()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / shortcuts / package
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        ]

import inspect
from functools import wraps

import six

from .. import init
from ..adapters import adapter_hub
from ..router import router_hub
from ..rendering.decorators import renderable
from ..utils.decorators import view


def expose(fn):
    '''``__all__`` registry helper.'''

    __all__.append(fn.__name__)
    return fn


# Expose convenient aliases.
expose(renderable)
expose(view)

make_app = adapter_hub.make_app
expose(make_app)

boot, inject_app = init.boot, init.inject_app
expose(boot)
expose(inject_app)

# Special case: the former load_all() got renamed to boot(), we need to
# preserve its name for compatibility (actually that's because I'm too lazy
# to rename all these occurences in my apps...)
load_all = init.boot

# NOTE: This str() call is correct as it IS the 'string' type recognized by
# the IMPORT_STAR thing, and it isn't that easy to make one using six.
__all__.append(str('load_all'))


@expose
def jsonview(fn):
    '''Short for doing ``@renderable('json')`` and ``@view`` in a row.'''

    return renderable('json')(view(fn))


def _transform_view_name(name):
    tmpname = name[:-5] if name.endswith('_view') else name
    return tmpname.replace('_', '-')


@expose
def http(name=None):
    '''Register a view for the HTTP router.

    If ``name`` is given as a string, this is a convenient form of
    ``@router_hub.endpoint('http', name)``. If ``name`` is not given
    (``@http()``), or if the decorator is used directly (``@http``), view
    name is derived from the function name by removing a ``_view`` suffix
    if present, and then replacing all underscores with hyphens.

    '''

    def _decorator_(fn):
        view_name = name or _transform_view_name(fn.__name__)
        return router_hub.endpoint('http', view_name)(fn)

    if callable(name):
        # Quick and dirty check to see if name is actually a function
        # (the @http case).
        fn, name = name, None
        return _decorator_(fn)

    return _decorator_



# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / rendering / decorator
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Decorator interface
~~~~~~~~~~~~~~~~~~~

'''

from __future__ import unicode_literals, division

__all__ = [
        'renderable',
        ]

import six

from ..helpers.annotation import annotate, get_annotation


def renderable(handler, *args, **kwargs):
    '''Marks an object as "renderable" by a certain handler, recording the
    render handler and associated instantiation parameter(s) in the object's
    attribute.

    :param handler: Name of the render handler.

    .. warning::

        The attribute used to store rendering information is currently
        ``_weiyu_rendering_``; this is an *implementation detail*\ .
        **Never** touch it outside the rendering engine proper as the exact
        detail of rendering attribute can be possibly radically different
        across versions.

    '''

    handler = six.text_type(handler)

    def _decorator_(thing):
        try:
            render_info = get_annotation(thing, 'rendering')
        except AttributeError:
            render_info = {}
        except KeyError:
            render_info = {}

        render_info[handler] = (args, kwargs, )

        annotate(thing, rendering=render_info)
        return thing
    return _decorator_


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / rendering / base operations
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Base classes for rendering
~~~~~~~~~~~~~~~~~~~~~~~~~~

This module provides the basic object rendering infrastructure, with
both pre-rendering and post-rendering hooks.

'''

from __future__ import unicode_literals, division

__all__ = ['RenderContext', 'Renderable', ]

import six

from .exc import RenderingError


class RenderContext(dict):
    '''Render context class.

    Currently this is just another name for ``dict``; may restrict type of
    keys to ``unicode`` only in the future when there is need.

    '''

    pass


def is_render_context(obj):
    return issubclass(type(obj), RenderContext)


class Renderable(object):
    '''Base class for any ``weiyu``-renderable object.

    It supports two distinct sets of pre- and post-rendering hooks.
    The separated sets of hooks seem to be more intuitive than Django's
    middleware system, which actually is what inspired this design.

    The pre-processor or post-processor can be any callable. Pre-processors
    are called with the result object and a ``RenderContext`` instance, and
    are expected to return a tuple ``(proceed, new_result, new_context, )``.
    If  ``proceed == False``, the actual rendering will be skipped and
    post-processor hooks will be invoked. The rendering result will be set
    to ``new_result`` in this case.

    The post-processors work almost the same way; they
    are invoked with two parameters, the first being the newest
    intermediate result of rendering, the second the final render context
    object. Their return values are interpreted in a similar fashion,
    *replacing the whole result string* when the returned value's type is a
    subclass of ``unicode``.

    '''

    def __init__(self, pre_hooks=None, post_hooks=None):
        self.pre_hooks = pre_hooks if pre_hooks is not None else []
        self.post_hooks = post_hooks if post_hooks is not None else []

    def _do_render(self, result, context):
        raise NotImplementedError

    def render(self, result=None, context=None):
        result = {} if result is None else result
        context = RenderContext() if context is None else context

        if not is_render_context(context):
            raise RenderingError(
                    'class of context must be subclass of RenderContext'
                    )

        skip_render, rendered = False, None
        cooked_result, cooked_ctx = result, context

        # Execute the preprocess hooks.
        for hook in self.pre_hooks:
            proceed, new_result, new_ctx = hook(cooked_result, cooked_ctx)

            # context
            if not is_render_context(new_ctx):
                raise RenderingError(
                        'new context not subclass of RenderContext'
                        )

            # update the result and context
            cooked_result, cooked_ctx = new_result, new_ctx

            # skip rendering?
            if not proceed:
                skip_render, rendered = True, cooked_result
                break

        # do the actual render if none of the preprocess hooks demanded early
        # return
        if not skip_render:
            rendered = self._do_render(cooked_result, cooked_ctx)

        # Execute the postprocess hooks, w/ the result string and context.
        for hook in self.post_hooks:
            new_rendered = hook(rendered, cooked_ctx)

            if new_rendered is None:
                # skip to the next one
                continue
            elif issubclass(type(new_rendered), six.text_type):
                # update (replace) rendered string
                rendered = new_rendered
            else:
                raise RenderingError(
                        "invalid post-process hook return type: got '%s'"
                        % str(type(new_rendered))
                        )

        return rendered


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / db drivers / base class
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

import abc

import six


class BaseDriver(six.with_metaclass(abc.ABCMeta)):
    '''Baseclass for database drivers.'''

    @abc.abstractmethod
    def start(self):
        '''Prepare for database operations.'''

        pass

    @abc.abstractmethod
    def finish(self):
        '''Be done with the database.

        The connection may be returned to some kind of pools, or closed.

        '''

        pass

    def __call__(self, bucket):
        '''Get a database connection for the selected bucket of data.

        This connection is driver-specific, for the developers to fully
        leverage the details of underlying database infrastructure.

        '''

        return DBOperationContext(self, bucket)

    @abc.abstractmethod
    def get_bucket(self, bucket):
        '''Get a driver-specific bucket/collection/table object to operate
        on.

        '''

        pass


class DBOperationContext(object):
    def __init__(self, driver, bucket):
        self.driver, self.bucket = driver, bucket

    def __enter__(self):
        self.driver.start()
        return self.driver.get_bucket(self.bucket)

    def __exit__(self, exc_type, exc_value, traceback):
        self.driver.finish()

    def __repr__(self):
        return b'<DBOperationContext: driver=%s, bucket=%s>' % (
                repr(self.driver),
                self.bucket,
                )


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / async / package
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'async_hub',
        ]

from ..helpers.hub import BaseHub
from ..registry.classes import UnicodeRegistry

FLAVOR_KEY, NS_KEY = 'flavors', 'namespaces'


class AsyncHub(BaseHub):
    registry_name = 'weiyu.async'
    registry_class = UnicodeRegistry
    handlers_key = FLAVOR_KEY  # XXX currently unused

    def __init__(self):
        super(AsyncHub, self).__init__()

        if NS_KEY not in self._reg:
            self._ns = self._reg[NS_KEY] = {}

    def register_ns(self, flavor, name):
        def _decorator_(thing):
            if flavor not in self._ns:
                self._ns[flavor] = {}

            flavor_ns = self._ns[flavor]
            if name in flavor_ns:
                raise ValueError(
                        "namespace '%s' already registered" % (
                            name,
                            )
                        )

            flavor_ns[name] = thing
            return thing
        return _decorator_

    def get_namespaces(self, flavor):
        return self._ns[flavor]


async_hub = AsyncHub()

# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import unicode_literals, division

# distribute and setuptools have merged since setuptools-0.7,
# so no more distribute_setup stuff here.
# Rest of the script should continue to work without modification.
from setuptools import setup
from setuptools import find_packages

from weiyu.__version__ import VERSION_STR


setup(
        name='weiyu',
        version=VERSION_STR.split(' ')[0],  # strip out the possible VCS commit
        description='Yet another Python Web framework, as modular and configurable as possible',
        author='Wang Xuerui',
        author_email='idontknw.wang+pypi@gmail.com',
        license='GPLv3+',
        url='https://github.com/xen0n/weiyu/',
        download_url='https://github.com/xen0n/weiyu/',
        install_requires=('decorator>=3.4.0', 'ply>=3.4', 'PyYAML>=3.10', 'six>=1.3.0', ),
        extras_require={
            'beaker': ['Beaker>=1.6.3', ],
            'mako': ['Mako>=0.7.1', ],
            'mongodb': ['pymongo>=2.2.1', ],
            'riak': ['riak>=2.0.0', ],
            'cherrypy': ['cherrypy>=3.2', ],
            'tornado': ['tornado>=2.4', ],
            'socketio': ['gevent-socketio>=0.3.5-rc2', ],
            'celery': ['celery>3.0', ],
            'memcached': ['python-memcached>=1.48', ],
            'redis': ['redis>=2.7.0', ],
            'hiredis': ['hiredis>=0.1.1', ],
            'ujson': ['ujson>=1.19', ],
            'yaml': [],
            'ghwebhook': ['ipaddr>=2.1.10', ],
            'scss': ['pyScss>=1.1.5', ],
            },
        packages=find_packages(exclude=['examples', 'tests']),
        classifiers=[
            'Development Status :: 2 - Pre-Alpha',
            'Environment :: Web Environment',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.2',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: Implementation :: CPython',
            'Programming Language :: Python :: Implementation :: PyPy',
            'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
            'Topic :: Internet :: WWW/HTTP :: WSGI',
            'Topic :: Software Development :: Libraries :: Application Frameworks',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        )


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / routing config / grammar
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division, print_function

__all__ = [
        'parser',
        ]

import re
import json
from collections import OrderedDict

import ply.yacc as yacc

from .wrlex import WRLexer
tokens = WRLexer.tokens


# general rule
def p_routedef(p):
    '''routedef : attriblist
                | attriblist targetlist
                | attriblist targetlist EOF'''
    attriblist = p[1]
    has_targets = len(p) > 2

    if len(attriblist) == 1:
        p[0] = p[1] + p[2] if has_targets else p[1]
    else:
        # wrap the attribs so that they always occupy 1 position
        p[0] = [p[1]] + p[2] if has_targets else [p[1]]
    #print 'new routedef %s' % (repr(p[0]), )


# ATTRIB list
def p_attriblist_list(p):
    'attriblist : attriblist attribdef NEWLINE'
    p[0] = p[1] + [p[2]]
    #print 'new attriblist:', p[0]


def p_attriblist_def(p):
    'attriblist : attribdef NEWLINE'
    p[0] = [p[1]]
    #print 'new attriblist:', p[0]


def p_attribdef(p):
    'attribdef : ATTRIB'
    p[0] = p[1][2:]
    #print 'ATTRIB: %s' % (repr(p[0]), )


# route target list
def p_targetlist_list(p):
    'targetlist : targetlist targetdeflf'
    p[0] = p[1] + [p[2]]
    #print 'new targetlist: %s' % (repr(p[0]), )


def p_targetlist_def(p):
    'targetlist : targetdeflf'
    p[0] = [p[1]]
    #print 'new targetlist: %s' % (repr(p[0]), )


def p_targetdef_lf(p):
    '''targetdeflf : targetdef
                   | targetdef NEWLINE'''
    p[0] = p[1]


def p_targetdef_simple(p):
    'targetdef : simpletgt'
    p[0] = p[1]


def p_targetdef_router(p):
    'targetdef : subrouter'
    p[0] = p[1]


# regular routing target
def p_simpletgt_simple(p):
    'simpletgt : pattern endpoint renderer'
    p[0] = [p[1], p[2], p[3], ]
    #print 'SIMPLETGT: %s' % (repr(p[0]), )


def p_simpletgt_defaultrenderer(p):
    'simpletgt : pattern endpoint'
    p[0] = [p[1], p[2], {'render_in': 'inherit', }, ]


def p_simpletgt_withextras(p):
    'simpletgt : pattern endpoint renderer extras'
    renderer = p[3]
    extras = p[4]

    if renderer is None:
        p[0] = [p[1], p[2], p[4], ]
    else:
        # renderer is a dictionary
        # NOTE: assertion that the extra data is also a dictionary
        new_extras = renderer.copy()
        new_extras.update(extras)
        p[0] = [p[1], p[2], new_extras, ]


# sub router declaration
def p_subrouter_simple(p):
    'subrouter : pattern COLON INDENT routedef cdedent'
    p[0] = [p[1], p[4], None, ]


def p_subrouter_withextras(p):
    'subrouter : pattern extras COLON INDENT routedef cdedent'
    p[0] = [p[1], p[5], p[2], ]
    #print 'SUBROUTER: %s' % (repr(p[0]), )


# primitives
def p_pattern(p):
    'pattern : LITERAL'
    p[0] = p[1]


def p_endpoint(p):
    'endpoint : LITERAL'
    p[0] = p[1]


def p_renderer(p):
    'renderer : LITERAL'
    p[0] = None if p[1] == 'null' else {'render_in': p[1]}


def p_extras(p):
    'extras : LITERAL'
    p[0] = json.loads(p[1])


# helpers
def p_combined_dedent(p):
    '''cdedent : DEDENT
               | NEWLINE DEDENT'''
    pass


def p_error(tok):
    import sys

    # TODO: proper error recovery
    print('*** Error: %s' % (repr(tok), ), file=sys.stderr)


# don't write parse table, the overhead is negligible when compared to
# the total uptime of web servers
parser = yacc.yacc(debug=0, write_tables=0)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / helpers / module hub management
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

'''
Dispatch hub module
~~~~~~~~~~~~~~~~~~~

'''

from __future__ import unicode_literals, division

__all__ = ['BaseHub', ]

from functools import partial

from ..registry.provider import request


class BaseHub(object):
    '''A module-level hub used to register handlers for the types of data it
    can process. It works by keeping a type-to-handler mapping ``dict`` in a
    registry; as a side effect, handlers can be added, modified or deleted
    on the fly, during system operation.

    The class is meant for usage in package-level ``__init__.py`` files::

        # in one of your __init__.py files:

        # best to only "expose" the Hub instance
        __all__ = ['Hub', ]

        from weiyu.helpers.hub import BaseHub
        from weiyu.registry.classes import UnicodeRegistry


        class SpamEggHub(BaseHub):
            # specify several required properties
            registry_name = 'sketch'
            registry_class = UnicodeRegistry
            handlers_key = 'handlers'


        # instantiate a single Hub instance
        Hub = SpamEggHub()

    Then you can register your functions like this::

        # in spam.py, which is inside the abovementioned package

        from . import Hub


        # handler function
        @Hub.register_handler('spam')
        def spam_handler(*args, **kwargs):
            pass


        # do similar things for the other .py files containing handlers to
        # be registered

    Just be sure to have ``import``\ ed the module before operation starts;
    otherwise the handler mapping is not there, for obvious reasons. This
    may get fixed in some later revisions.

    The class attributes that must be present in subclasses is described
    below:

    :attr registry_name: name of the registry used to hold the
      type-to-handler mapping;
    :attr registry_class: the proper registry class to use, if the registry
      named by ``registry_name`` does not exist at the time of hub creation;
    :attr handlers_key: key which the type-to-handler mapping resides in.

    To avoid excessive attribute lookups when retrieving reference to the
    internal mapping, the mapping ``dict`` is cached. If it turns out that
    dynamically substituting the handler mapping object is necessary, you
    need to ensure that :meth:`refresh_handler_map` is called right after
    applying the object replacement, and before any potential call to get a
    handler. As long as you don't totally replace the mapping object (which
    is almost always the case), calling this method is NOT needed and
    meaningless, as it will read back the same reference.

    '''

    def __init__(self):
        cls = self.__class__
        self._reg = request(
                cls.registry_name,
                autocreate=True,
                nodup=False,
                klass=cls.registry_class,
                )

        if cls.handlers_key not in self._reg:
            self._reg[cls.handlers_key] = {}

        # this is to reduce attribute lookups
        self.refresh_handler_map()

    def refresh_handler_map(self):
        '''Re-reads in-registry reference to the type-to-handler mapping.

        This method exists only in case the mapping needs to be pointed to
        a brand-new object; it should not be called in most cases.

        '''

        self._handlers = self._reg[self.__class__.handlers_key]

    def register_handler(self, typ):
        '''Decorator to register a handler for type ``typ``; the decorated
        object is returned untouched.

        In case a handler for ``typ`` is already registered, the previous
        reference is replaced by the new one.

        '''

        def _decorator_(thing):
            # NOTE: Doesn't work with FunctionValueRegistry
            self._handlers[typ] = thing
            return thing
        return _decorator_

    def remove_handler(self, typ):
        '''Removes the handler associated with type ``typ``; does nothing
        if the handler does not exist.

        '''

        try:
            self._handlers.pop(typ)
        except KeyError:
            pass

    def get_handler(self, typ):
        '''Get the handler associated with type ``typ``.

        .. note::
            Exceptions are intentionally not caught in this method, you
            may have to explicitly deal with the possible cases.

        '''

        # NOTE: Exception is intentionally propagated, a KeyError for
        # unregistered handler is nice...
        # Also, handlers' signatures all have a 'hub' argument as their
        # first parameter, so we'd curry the hub itself into the resulting
        # callable.
        return partial(self._handlers[typ], self)

    def do_handling(self, typ, *args, **kwargs):
        '''Get the handler, and additionally do the work of invoking the
        handler with the proper argument specified, in one step. This is
        a shortcut method.

        '''

        return self.get_handler(typ)(*args, **kwargs)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / helpers / module importer
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'ModProber',
        ]

import importlib


class ModProber(object):
    def __init__(self, package, name_template, package_map=None):
        self.package = package
        self.name_template = name_template
        self.package_map = package_map or {}

    def get_relative_path(self, name):
        assert '.' not in name
        return '.' + self.name_template % (self.package_map.get(name, name), )

    def modprobe(self, name):
        return importlib.import_module(
                self.get_relative_path(name),
                self.package,
                )


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / rendering / package
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'render_hub',
        ]

from ..helpers.hub import BaseHub
from ..helpers.modprober import ModProber
from ..helpers.annotation import get_annotation
from ..registry.classes import UnicodeRegistry

from .exc import RenderingError
from .base import RenderContext

PROBER = ModProber('weiyu.rendering', '%srenderer')


class RenderHub(BaseHub):
    registry_name = 'weiyu.rendering'
    registry_class = UnicodeRegistry
    handlers_key = 'handlers'

    # template thing
    def get_template(self, typ, name=None, *args, **kwargs):
        if typ not in self._handlers:
            # Do delayed loading of renderer module
            PROBER.modprobe(typ)

        # name=None is here so those templateless renderers (like JSON)
        # can be used w/o a dummy parameter
        return self.do_handling(typ, name, *args, **kwargs)

    def render_view(self, renderable_fn, result, context, typ):
        render_info = get_annotation(renderable_fn, 'rendering')

        if typ not in render_info:
            # this format is not supported by view
            # TODO: maybe specialize this exception's type
            raise RenderingError(
                    'This format ("%s") is not supported by view' % typ
                    )

        ctx = RenderContext(context)
        handler_args, handler_kwargs = render_info[typ]
        tmpl = self.get_template(typ, *handler_args, **handler_kwargs)
        return tmpl.render(result, ctx)


render_hub = RenderHub()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / session handling / package
#
# Copyright (C) 2012-2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'session_hub',
        ]

from ..helpers.hub import BaseHub
from ..helpers.modprober import ModProber
from ..registry.classes import UnicodeRegistry
from ..registry.provider import request

PROBER = ModProber('weiyu.session', '%sbackend')


class SessionHub(BaseHub):
    registry_name = 'weiyu.session'
    registry_class = UnicodeRegistry
    handlers_key = 'backends'

    def do_handling(self, typ, *args, **kwargs):
        # load the selected session backend
        PROBER.modprobe(typ)

        return super(SessionHub, self).do_handling(typ, *args, **kwargs)


session_hub = SessionHub()


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / utilities / decorators
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
        'view',
        'only_methods',
        ]

import decorator

from ..reflex.classes import ReflexResponse
from ..helpers.annotation import annotate


def _view_func_(fn, request, *args, **kwargs):
    status, content, context = fn(request, *args, **kwargs)
    return ReflexResponse(
            status,
            content,
            context,
            request,
            )


def view(fn):
    '''View decorator to avoid having to
    ``from weiyu.reflex.classes import ReflexResponse`` everywhere.

    '''

    return decorator.decorator(_view_func_, fn)


def only_methods(methods=None):
    '''Decorator to only allow certain types of HTTP methods.

    Defaults to ``GET`` only.

    Must occur *after* the ``@view`` decorator.

    '''

    methods = ['GET', ] if methods is None else methods

    def _decorator_(fn):
        # Although the actual binding of methods list occurs in
        # closure scope, we keep a reference of it in annotation
        # for clarity.
        annotate(fn, allowed_methods=methods)

        def _wrapped_(fn, request, *args, **kwargs):
            if request.method not in methods:
                # method not allowed
                # set an Allow header
                allowed_hdr = ', '.join(methods).encode('utf-8')
                return (
                        405,
                        {},
                        {'allowed_methods': allowed_hdr, },
                        )

            return fn(request, *args, **kwargs)
        return decorator.decorator(_wrapped_, fn)
    return _decorator_


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / cache drivers / memcached driver
#
# Copyright (C) 2013 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

import memcache

from .. import cache_hub
from .baseclass import BaseCache


class MemcachedCache(BaseCache):
    def __init__(self, servers):
        super(MemcachedCache, self).__init__()
        self._conn = memcache.Client(servers)

    def get(self, k):
        return self._conn.get(k)

    def set(self, k, v, timeout=None):
        return self._conn.set(k, v)

    def delete(self, k):
        return self._conn.delete(k)


@cache_hub.register_handler('memcached')
def memcached_handler(hub, opts):
    servers = opts['servers']

    return MemcachedCache(servers)


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# weiyu / router / exact-matching router
#
# Copyright (C) 2012 Wang Xuerui <idontknw.wang-at-gmail-dot-com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals, division

__all__ = [
            'ExactRouterTarget',
            'ExactRouter',
            ]

import re

from ..helpers.misc import smartstr
from . import router_hub
from .base import *


class ExactRouterTarget(RouterTargetBase):
    '''
    This target cannot extract any arguments from query strings; useful
    when a large tree of, say, API endpoints are to be exposed, and the
    parameters are not passed through the query string. It's supposed
    to be faster than regex-based lookups in such static cases.

    .. note::
        No separator is automatically removed from the matched query
        string for nested lookups; you should specify the separator such
        as '/' yourself in the target pattern.

    '''

    def __init__(self, pattern, target, extra_data=None, router=None):
        super(ExactRouterTarget, self).__init__(target, extra_data, router)

        self.pattern = smartstr(pattern)
        self._pat_len = len(self.pattern)

    def check(self, querystr, prev_args, prev_kwargs):
        match = querystr.startswith(self.pattern)

        if not match:
            return (STATUS_NOROUTE, None, None, None, )

        # support for nested routing
        if not self.target_is_router:
            return (STATUS_REACHED, prev_args, prev_kwargs, None, )

        # target is router, forward
        # since the RE matches at the beginning, safely assume its span()
        # starts at 0.
        # now, extract the substring for further routing...
        new_qs = querystr[self._pat_len:]

        # and indicate this status.
        return (STATUS_FORWARD, prev_args, prev_kwargs, new_qs, )

    def _get_reverse_pattern(self):
        return (self.pattern, set(), )


@router_hub.register_router_class('exact')
class ExactRouter(RouterBase):
    target_type = ExactRouterTarget


# vim:set ai et ts=4 sw=4 sts=4 fenc=utf-8:
